"""Switcher water heater unofficial API and bridge.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

name = "aioswitcher"

__all__ = ["api", "bridge", "devices", "schedules", "tools"]
